section			:	T_POP_SECTION string ',' sectiontype
					{ out_NewSection($2,$4); }
		|	T_POP_SECTION string ',' sectiontype '[' const ']'
					{
						if( $6>=0 && $6<0x10000 )
							out_NewAbsSection($2,$4,$6,-1);
						else
							yyerror( "Address must be 16-bit" );
					}
		|	T_POP_SECTION string ',' sectiontype ',' T_OP_BANK '[' const ']'
					{
						if( $4==SECT_CODE )
						{
							if( $8>=1 && $8<=255 )
								out_NewAbsSection($2,$4,-1,$8);
							else
								yyerror( "BANK value out of range" );
						}
						else
							yyerror( "BANK only allowed for CODE/DATA" );
					}
		|	T_POP_SECTION string ',' sectiontype '[' const ']' ',' T_OP_BANK '[' const ']'
					{
						if( $4==SECT_CODE )
						{
							if( $6>=0 && $6<0x10000 )
							{
								if( $11>=1 && $11<=255 )
									out_NewAbsSection($2,$4,$6,$11);
								else
									yyerror( "BANK value out of range" );
							}
							else
								yyerror( "Address must be 16-bit" );
						}
						else
							yyerror( "BANK only allowed for CODE/DATA" );
					}
;

sectiontype		:	T_SECT_BSS	{ $$=SECT_BSS; }
			|	T_SECT_VRAM		{ $$=SECT_VRAM; 	}
			|	T_SECT_CODE		{ $$=SECT_CODE; 	}
			|	T_SECT_SAVESSCREEN	{ $$=SECT_SAVESCREEN;	}
			|	T_SECT_STATVARS		{ $$=SECT_STATVARS;	}
			|	T_SECT_TEXTSHADOW	{ $$=SECT_TEXTSHADOW;	}
			|	T_SECT_PLOTSCREEN	{ $$=SECT_PLOTSCREEN;	}
;


cpu_command		:	z80_adc
			|	z80_add
			|	z80_and
			|	z80_bit
			|	z80_call
			|	z80_ccf
			|	z80_cp
			|	z80_cpd
			|	z80_cpdr
			|	z80_cpi
			|	z80_cpir
			|	z80_cpl
			|	z80_daa
			|	z80_dec
			|	z80_di
			|	z80_djnz
			|	z80_ei
			|	z80_ex
			|	z80_exx
			|	z80_halt
			|	z80_im
			|	z80_in
			|	z80_ind
			|	z80_indr
			|	z80_ini
			|	z80_inir
			|	z80_inc
			|	z80_jp
			|	z80_jr
			|	z80_ld
			|	z80_ldd
			|	z80_lddr
			|	z80_ldi
			|	z80_ldir
			|	z80_neg
			|	z80_nop
			|	z80_or
			|	z80_out
			|	z80_outd
			|	z80_otdr
			|	z80_outi
			|	z80_otir
			|	z80_pop
			|	z80_push
			|	z80_res
			|	z80_ret
			|	z80_reti
			|	z80_retn
			|	z80_rl
			|	z80_rla
			|	z80_rlc
			|	z80_rlca
			|	z80_rld
			|	z80_rr
			|	z80_rra
			|	z80_rrc
			|	z80_rrca
			|	z80_rrd
			|	z80_rst
			|	z80_sbc
			|	z80_scf
			|	z80_set
			|	z80_sla
			|	z80_sra
			|	z80_srl
			|	z80_sub
			|	z80_xor
;

z80_adc			:	T_Z80_ADC op_a_n
					{ out_AbsByte(0xCE); out_RelByte(&$2); }
			|	T_Z80_ADC op_a_r
					{ out_AbsByte(0x88|$2); }
			|	T_Z80_ADC T_MODE_HL comma reg_ss
					{ out_AbsByte(0xED); out_AbsByte(0x4A | ($4 << 4)); }
			|	T_Z80_ADC T_MODE_A comma op_indexed
					{ out_AbsByte($4.nUserData); out_AbsByte(0x8E); out_RelByte(&$4); }
;

z80_add			:	T_Z80_ADD op_a_n	{ out_AbsByte(0xC6); out_RelByte(&$2); }
			|	T_Z80_ADD op_a_r	{ out_AbsByte(0x80|$2); }
			|	T_Z80_ADD op_hl_ss	{ out_AbsByte(0x09|($2<<4)); }
			|	T_Z80_ADD T_MODE_A comma op_indexed
					{ out_AbsByte($4.nUserData); out_AbsByte(0x86); out_RelByte(&$4); }
			|	T_Z80_ADD reg_i comma reg_ss
					{
						if($4 == REG_HL)
							yyerror("Source must not be HL!");
						else
							{ out_AbsByte($2); out_AbsByte(0x09 | ($4 << 4)); }
					}
			|	T_Z80_ADD reg_i comma reg_i
					{
						if($2 != $4)
							yyerror("Invalid operands for add!");
						else
							{ out_AbsByte($2); out_AbsByte(0x29); }
					}

;

z80_and			:	T_Z80_AND op_a_n	{ out_AbsByte(0xE6); out_RelByte(&$2); }
			|	T_Z80_AND op_a_r	{ out_AbsByte(0xA0|$2); }
			|	T_Z80_AND op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xA6); out_RelByte(&$2); }
;

z80_bit			:	T_Z80_BIT const_3bit comma reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x40|($2<<3)|$4); }
			|	T_Z80_BIT const_3bit comma op_indexed
					{ out_AbsByte($4.nUserData); out_AbsByte(0xCB); out_RelByte(&$4); out_AbsByte(0x46 | ($2 << 3)); }
;

z80_call		:	T_Z80_CALL const_16bit
					{ out_AbsByte(0xCD); out_RelWord(&$2); }
			|	T_Z80_CALL ccode comma const_16bit
					{ out_AbsByte(0xC4|($2<<3)); out_RelWord(&$4); }
;

z80_ccf			:	T_Z80_CCF
					{ out_AbsByte(0x3F); }
;

z80_cp			:	T_Z80_CP op_a_n	{ out_AbsByte(0xFE); out_RelByte(&$2); }
			|	T_Z80_CP op_a_r	{ out_AbsByte(0xB8|$2); }
			|	T_Z80_CP op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xBE); out_RelByte(&$2); }
;

z80_cpd			:	T_Z80_CPD
					{ out_AbsByte(0xED); out_AbsByte(0xA9); }
;

z80_cpdr		:	T_Z80_CPDR
					{ out_AbsByte(0xED); out_AbsByte(0xB9); }
;

z80_cpi			:	T_Z80_CPI
					{ out_AbsByte(0xED); out_AbsByte(0xA1); }
;

z80_cpir		:	T_Z80_CPIR
					{ out_AbsByte(0xED); out_AbsByte(0xB1); }
;

z80_cpl			:	T_Z80_CPL { out_AbsByte(0x2F); }
;

z80_daa			:	T_Z80_DAA { out_AbsByte(0x27); }
;

z80_dec			:	T_Z80_DEC reg_r
					{ out_AbsByte(0x05|($2<<3)); }
			|	T_Z80_DEC reg_ss
					{ out_AbsByte(0x0B|($2<<4)); }
			|	T_Z80_DEC reg_i
					{ out_AbsByte($2); out_AbsByte(0x2B); }						
			|	T_Z80_DEC op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0x35); out_RelByte(&$2); }
;

z80_di			:	T_Z80_DI
					{ out_AbsByte(0xF3); }
;

z80_djnz			:	T_Z80_DJNZ const_PCrel
					{ out_AbsByte(0x10); out_PCRelByte(&$2); }
;

z80_ei			:	T_Z80_EI
					{ out_AbsByte(0xFB); }
;

z80_ex			:	T_Z80_EX T_MODE_HL comma T_MODE_SP_IND
					{ out_AbsByte(0xE3); }
			|	T_Z80_EX T_MODE_SP_IND comma T_MODE_HL
					{ out_AbsByte(0xE3); }
			|	T_Z80_EX T_MODE_AF comma T_MODE_AFP
					{ out_AbsByte(0x08); }
			|	T_Z80_EX T_MODE_AFP comma T_MODE_AF
					{ out_AbsByte(0x08); }
			|	T_Z80_EX T_MODE_DE comma T_MODE_HL
					{ out_AbsByte(0xEB); }
			|	T_Z80_EX T_MODE_HL comma T_MODE_DE
					{ out_AbsByte(0xEB); }
			|	T_Z80_EX T_MODE_SP_IND comma T_MODE_IX
					{ out_AbsByte(0xDD); out_AbsByte(0xE3); }
			|	T_Z80_EX T_MODE_IX comma T_MODE_SP_IND
					{ out_AbsByte(0xDD); out_AbsByte(0xE3); }
			|	T_Z80_EX T_MODE_SP_IND comma T_MODE_IY
					{ out_AbsByte(0xFD); out_AbsByte(0xE3); }
			|	T_Z80_EX T_MODE_IY comma T_MODE_SP_IND
					{ out_AbsByte(0xFD); out_AbsByte(0xE3); }
;

z80_exx			:	T_Z80_EXX
					{ out_AbsByte(0xD9); }
;

z80_halt			:	T_Z80_HALT
					{ out_AbsByte(0x76); out_AbsByte(0x00); }
;

z80_im			:	T_Z80_IM const_8bit
					{
					if( rpn_isReloc(&$2) )
						yyerror( "Value of x for 'IM x' must be constant!" );
					else if( $2.nVal == 0 )
						{ out_AbsByte(0xED); out_AbsByte(0x46); }
					else if( $2.nVal == 1 )
						{ out_AbsByte(0xED); out_AbsByte(0x56); }
					else if( $2.nVal == 2 )
						{ out_AbsByte(0xED); out_AbsByte(0x5E); }
					else
						yyerror( "Invalid value for x in 'IM x'!");
					}

;

z80_in			:	T_Z80_IN reg_r comma '[' const_8bit ']'
					{
						if( $2==REG_A )
							{ out_AbsByte(0xDB); out_RelByte(&$5); }
						else
						{
							yyerror( "Destination operand must be A" );
						}
					}
			|	T_Z80_IN reg_r comma T_MODE_C_IND
					{ out_AbsByte(0xED); out_AbsByte(0x40|($2<<3)); }
;


z80_inc			:	T_Z80_INC reg_r
					{ out_AbsByte(0x04|($2<<3)); }
			|	T_Z80_INC reg_ss
					{ out_AbsByte(0x03|($2<<4)); }
			|	T_Z80_INC reg_i
					{ out_AbsByte($2); out_AbsByte(0x23); }
			|	T_Z80_INC op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0x34); out_RelByte(&$2); }
;

z80_ind			:	T_Z80_IND
					{ out_AbsByte(0xED); out_AbsByte(0xAA); }
;

z80_indr		:	T_Z80_INDR
					{ out_AbsByte(0xED); out_AbsByte(0xBA); }
;

z80_ini			:	T_Z80_INI
					{ out_AbsByte(0xED); out_AbsByte(0xA2); }
;

z80_inir		:	T_Z80_INIR
					{ out_AbsByte(0xED); out_AbsByte(0xB2); }
;

z80_jp			:	T_Z80_JP const_16bit
					{ out_AbsByte(0xC3); out_RelWord(&$2); }
			|	T_Z80_JP ccode comma const_16bit
					{ out_AbsByte(0xC2|($2<<3)); out_RelWord(&$4); }
			|	T_Z80_JP T_MODE_HL_IND
					{ out_AbsByte(0xE9); }
			|	T_Z80_JP '[' reg_i ']'
					{ out_AbsByte($3); out_AbsByte(0xE9); }
;

z80_jr			:	T_Z80_JR const_PCrel
					{ out_AbsByte(0x18); out_PCRelByte(&$2); }
			|	T_Z80_JR ccode comma const_PCrel
					{ out_AbsByte(0x20|($2<<3)); out_PCRelByte(&$4); }
;

z80_ldd			:	T_Z80_LDD
					{ out_AbsByte(0xED); out_AbsByte(0xA8); }
;

z80_lddr		:	T_Z80_LDDR
					{ out_AbsByte(0xED); out_AbsByte(0xB8); }
;

z80_ldi			:	T_Z80_LDI
					{ out_AbsByte(0xED); out_AbsByte(0xA0); }
;

z80_ldir		:	T_Z80_LDIR
					{ out_AbsByte(0xED); out_AbsByte(0xB0); }
;

z80_ld			:	z80_ld_mem
			|	z80_ld_rr
			|	z80_ld_hl
			|	z80_ld_sp
			|	z80_ld_r
			|	z80_ld_i
			|	z80_ld_a
			|	z80_ld_ss
			|	z80_ld_indexed
;

z80_ld_hl		:	T_Z80_LD T_MODE_HL comma const_16bit
					{ out_AbsByte(0x01|(REG_HL<<4)); out_RelWord(&$4) }
			|	T_Z80_LD T_MODE_HL comma op_mem_ind
					{ out_AbsByte(0xED); out_AbsByte(0x4B | (REG_HL << 4)); out_RelWord(&$4) }
;
z80_ld_sp		:	T_Z80_LD T_MODE_SP comma T_MODE_HL
					{ out_AbsByte(0xF9); }
			|	T_Z80_LD T_MODE_SP comma const_16bit
					{ out_AbsByte(0x01|(REG_SP<<4)); out_RelWord(&$4) }
			|	T_Z80_LD T_MODE_SP comma op_mem_ind
					{ out_AbsByte(0xED); out_AbsByte(0x4B | (REG_SP << 4)); out_RelWord(&$4) }
			|	T_Z80_LD T_MODE_SP comma reg_i
					{ out_AbsByte($4); out_AbsByte(0xF9); }
;

z80_ld_mem		:	T_Z80_LD op_mem_ind comma reg_ss
					{ out_AbsByte(0x0ED); out_AbsByte(0x43 | ($4 << 4)); out_RelWord(&$2); }
			|	T_Z80_LD op_mem_ind comma T_MODE_A
					{	out_AbsByte(0x32); out_RelWord(&$2); }
			|	T_Z80_LD op_mem_ind comma reg_i
					{ out_AbsByte($4); out_AbsByte(0x22); out_RelWord(&$2); }
;

z80_ld_rr		:	T_Z80_LD reg_rr comma T_MODE_A
					{ out_AbsByte(0x02|($2<<4)); }
;

z80_ld_r		:	T_Z80_LD reg_r comma const_8bit
					{ out_AbsByte(0x06|($2<<3)); out_RelByte(&$4); }
			|	T_Z80_LD reg_r comma reg_r
					{
						if( ($2==REG_HL_IND) && ($4==REG_HL_IND) )
						{
							yyerror( "LD [HL],[HL] not allowed" );
						}
						else
							out_AbsByte(0x40|($2<<3)|$4);
					}
			|	T_Z80_LD reg_r comma op_indexed
					{
						if($2 == REG_HL_IND)
							yyerror("Can not load into [HL] with indexed adressing");
						else
							{ out_AbsByte($4.nUserData); out_AbsByte(0x46 | ($2 << 3)); out_RelByte(&$4); }
					}
			|	T_Z80_LD reg_i comma op_mem_ind
					{ out_AbsByte($2); out_AbsByte(0x2A); out_RelWord(&$4); }
;

z80_ld_i		:	T_Z80_LD T_MODE_I comma T_MODE_A
					{ out_AbsByte(0xED); out_AbsByte(0x47); }
;

z80_ld_a		:	T_Z80_LD reg_r comma reg_rr
					{
						if( $2==REG_A )
							out_AbsByte(0x0A|($4<<4));
						else
						{
							yyerror( "Destination operand must be A" );
						}
					}
			|	T_Z80_LD reg_r comma op_mem_ind
					{
						if( $2==REG_A )
						{
							out_AbsByte(0x3A);
							out_RelWord(&$4);
						}
						else
						{
							yyerror( "Destination operand must be A" );
						}
					}
			|	T_Z80_LD reg_r comma T_MODE_I
					{
						if( $2==REG_A )
							{ out_AbsByte(0xED); out_AbsByte(0x57); }
						else
							yyerror( "Destination operand must be A" );
					}

;

z80_ld_ss		:	T_Z80_LD reg_ss comma const_16bit
					{ out_AbsByte(0x01|($2<<4)); out_RelWord(&$4) }
			|	T_Z80_LD reg_ss comma op_mem_ind
					{ out_AbsByte(0xED); out_AbsByte(0x4B | ($2 << 4)); out_RelWord(&$4) }
			|	T_Z80_LD reg_i comma const_16bit
					{ out_AbsByte($2); out_AbsByte(0x21); out_RelWord(&$4); }

;

z80_ld_indexed		:	T_Z80_LD op_indexed comma const_8bit
					{ out_AbsByte($2.nUserData); out_AbsByte(0x36); out_RelByte(&$2); out_RelByte(&$4); }
			|	T_Z80_LD op_indexed comma reg_r
					{
						if($4 == REG_HL_IND)
							yyerror("Can not load from [HL] with indexed adressing");
						else
							{ out_AbsByte($2.nUserData); out_AbsByte(0x70 | $4); out_RelByte(&$2); }
					}

z80_neg			:	T_Z80_NEG
					{ out_AbsByte(0xED); out_AbsByte(0x44); }

z80_nop			:	T_Z80_NOP
					{ out_AbsByte(0x00); }
;

z80_or			:	T_Z80_OR op_a_n
					{ out_AbsByte(0xF6); out_RelByte(&$2); }
			|	T_Z80_OR op_a_r
					{ out_AbsByte(0xB0|$2); }
			|	T_Z80_OR op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xB6); out_RelByte(&$2); }
;

z80_out			:	T_Z80_OUT '[' const_8bit ']' comma reg_r
					{
						if( $6==REG_A )
							{ out_AbsByte(0xD3); out_RelByte(&$3); }
						else
						{
							yyerror( "Source operand must be A" );
						}
					}
			|	T_Z80_OUT T_MODE_C_IND comma reg_r
					{ out_AbsByte(0xED); out_AbsByte(0x41|($4 << 3)); }
;

z80_outd		:	T_Z80_OUTD
					{ out_AbsByte(0xED); out_AbsByte(0xAB); }
;

z80_otdr		:	T_Z80_OTDR
					{ out_AbsByte(0xED); out_AbsByte(0xBB); }
;

z80_outi		:	T_Z80_OUTI
					{ out_AbsByte(0xED); out_AbsByte(0xA3); }
;

z80_otir		:	T_Z80_OTIR
					{ out_AbsByte(0xED); out_AbsByte(0xB3); }

z80_pop			:	T_Z80_POP reg_tt
					{ out_AbsByte(0xC1|($2<<4)); }
			|	T_Z80_POP reg_i
					{ out_AbsByte($2); out_AbsByte(0xE1); }
;

z80_push		:	T_Z80_PUSH reg_tt
					{ out_AbsByte(0xC5|($2<<4)); }
			|	T_Z80_PUSH reg_i
					{ out_AbsByte($2); out_AbsByte(0xE5); }					
;

z80_res			:	T_Z80_RES const_3bit comma reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x80|($2<<3)|$4); }
			|	T_Z80_RES const_3bit comma op_indexed
					{ out_AbsByte($4.nUserData); out_AbsByte(0xCB); out_RelByte(&$4); out_AbsByte(0x86 | ($2 << 3)); }
;

z80_ret			:	T_Z80_RET
					{ out_AbsByte(0xC9); }
			|	T_Z80_RET ccode
					{ out_AbsByte(0xC0|($2<<3)); }
;

z80_reti		:	T_Z80_RETI
					{ out_AbsByte(0xED); out_AbsByte(0x4D);}
;

z80_retn		:	T_Z80_RETN
					{ out_AbsByte(0xED); out_AbsByte(0x45);}
;

z80_rl			:	T_Z80_RL reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x10|$2); }
			|	T_Z80_RL op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xCB); out_RelByte(&$2); out_AbsByte(0x16); }
;

z80_rla			:	T_Z80_RLA
					{ out_AbsByte(0x17); }
;

z80_rlc			:	T_Z80_RLC reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x00|$2); }
			|	T_Z80_RLC op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xCB); out_RelByte(&$2); out_AbsByte(0x06); }
;

z80_rlca 		:	T_Z80_RLCA
					{ out_AbsByte(0x07); }
;

z80_rld			:	T_Z80_RLD
					{ out_AbsByte(0xED); out_AbsByte(0x6F); }
;

z80_rr			:	T_Z80_RR reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x18|$2); }
			|	T_Z80_RR op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xCB); out_RelByte(&$2); out_AbsByte(0x1E); }
;

z80_rra			:	T_Z80_RRA
					{ out_AbsByte(0x1F); }
;

z80_rrc			:	T_Z80_RRC reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x08|$2); }
			
			|	T_Z80_RRC op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xCB); out_RelByte(&$2); out_AbsByte(0x0E); }
;

z80_rrca 		:	T_Z80_RRCA
					{ out_AbsByte(0x0F); }
;

z80_rrd			:	T_Z80_RRD
					{ out_AbsByte(0xED); out_AbsByte(0x67); }
;

z80_rst 		:	T_Z80_RST const_8bit
					{
						if( rpn_isReloc(&$2) )
						{
							yyerror( "Address for RST must be absolute" );
						}
						else if( ($2.nVal&0x38)!=$2.nVal )
						{
							yyerror( "Invalid address for RST" );
						}
						else
							out_AbsByte(0xC7|$2.nVal);
					}
;

z80_sbc			:	T_Z80_SBC op_a_n
					{ out_AbsByte(0xDE); out_RelByte(&$2); }
			|	T_Z80_SBC op_a_r
					{ out_AbsByte(0x98|$2); }
			|	T_Z80_SBC T_MODE_HL comma reg_ss
					{ out_AbsByte(0xED); out_AbsByte(0x42 | ($4 << 4)); }
			|	T_Z80_SBC T_MODE_A comma op_indexed
					{ out_AbsByte($4.nUserData); out_AbsByte(0x9E); out_RelByte(&$4); }
;

z80_scf			:	T_Z80_SCF
					{ out_AbsByte(0x37); }
;

z80_set			:	T_POP_SET const_3bit comma reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0xC0|($2<<3)|$4); }
			|	T_POP_SET const_3bit comma op_indexed
					{ out_AbsByte($4.nUserData); out_AbsByte(0xCB); out_RelByte(&$4); out_AbsByte(0xC6 | ($2 << 3)); }
;

z80_sla			:	T_Z80_SLA reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x20|$2); }
			|	T_Z80_SLA op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xCB); out_RelByte(&$2); out_AbsByte(0x26); }
;

z80_sra			:	T_Z80_SRA reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x28|$2); }
			|	T_Z80_SRA op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xCB); out_RelByte(&$2); out_AbsByte(0x2E); }
;

z80_srl			:	T_Z80_SRL reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x38|$2); }
;

z80_sub			:	T_Z80_SUB op_a_n	{ out_AbsByte(0xD6); out_RelByte(&$2); }
			|	T_Z80_SUB op_a_r	{ out_AbsByte(0x90|$2); }
			|	T_Z80_SUB T_MODE_A comma op_indexed
					{ out_AbsByte($4.nUserData); out_AbsByte(0x96); out_RelByte(&$4); }
;

z80_xor			:	T_Z80_XOR op_a_n	{ out_AbsByte(0xEE); out_RelByte(&$2); }
			|	T_Z80_XOR op_a_r	{ out_AbsByte(0xA8|$2); }
			|	T_Z80_XOR op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xAE); out_RelByte(&$2); }
;

op_mem_ind		:	'[' const_16bit ']'	{ $$ = $2 }
;

op_indexed		:	'[' T_MODE_IX const_8bit ']' { $$ = $3; $$.nUserData = 0xDD; }
			|	'[' T_MODE_IY const_8bit ']' { $$ = $3; $$.nUserData = 0xFD; }
;

op_hl_ss 		:	reg_ss					{ $$ = $1 }
			|	T_MODE_HL comma reg_ss	{ $$ = $3 }
;

op_a_r			:	reg_r				{ $$ = $1 }
			|	T_MODE_A comma reg_r	{ $$ = $3 }
;

op_a_n			:	const_8bit				{ $$ = $1 }
			|	T_MODE_A comma const_8bit	{ $$ = $3 }
;

comma			:	','
;

ccode			:	T_CC_NZ		{ $$ = CC_NZ }
			|	T_CC_Z		{ $$ = CC_Z  }
			|	T_CC_NC		{ $$ = CC_NC }
			|	T_MODE_C	{ $$ = CC_C  }
			|	T_CC_PO		{ $$ = CC_PO }
			|	T_CC_PE		{ $$ = CC_PE }
			|	T_CC_P		{ $$ = CC_P  }
			|	T_CC_M		{ $$ = CC_M }
;

reg_r			:	T_MODE_B		{ $$ = REG_B }
			|	T_MODE_C		{ $$ = REG_C }
			|	T_MODE_D		{ $$ = REG_D }
			|	T_MODE_E		{ $$ = REG_E }
			|	T_MODE_H		{ $$ = REG_H }
			|	T_MODE_L		{ $$ = REG_L }
			|	T_MODE_HL_IND		{ $$ = REG_HL_IND }
			|	T_MODE_A		{ $$ = REG_A }
;

reg_i			:	T_MODE_IX		{ $$ = 0xDD }
			|	T_MODE_IY		{ $$ = 0xFD }
;

reg_tt			:	T_MODE_BC		{ $$ = REG_BC }
			|	T_MODE_DE		{ $$ = REG_DE }
			|	T_MODE_HL		{ $$ = REG_HL }
			|	T_MODE_AF		{ $$ = REG_AF }
;

reg_ss			:	T_MODE_BC		{ $$ = REG_BC }
			|	T_MODE_DE		{ $$ = REG_DE }
			|	T_MODE_HL		{ $$ = REG_HL }
			|	T_MODE_SP		{ $$ = REG_SP }
;

reg_rr			:	T_MODE_BC_IND		{ $$ = REG_BC_IND }
			|	T_MODE_DE_IND		{ $$ = REG_DE_IND }
;

%%