%{
#include "symbol.h"
#include "asm.h"
#include "output.h"
#include "mylink.h"
#include "fstack.h"
#include "mymath.h"
#include "rpn.h"
#include "main.h"
#include "lexer.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include	<ctype.h>

extern void * alloca(unsigned);

char	*tzNewMacro;
ULONG	ulNewMacroSize;

ULONG	symvaluetostring( char *dest, char *sym )
{
	if( sym_isString(sym) )
		strcpy( dest, sym_GetStringValue(sym) );
	else
		sprintf( dest, "$%lX", sym_GetConstantValue(sym) );

	return( strlen(dest) );
}

ULONG	str2int( char *s )
{
	ULONG r=0;
	while( *s )
	{
		r<<=8;
		r|=(UBYTE)(*s++);
	}
	return( r );
}

ULONG	isWhiteSpace( char s )
{
	return( s==' ' || s=='\t' || s=='\0' || s=='\n' );
}

ULONG	isRept( char *s )
{
	return( (strnicmp(s,"REPT",4)==0) && isWhiteSpace(*(s-1)) && isWhiteSpace(s[4]) );
}

ULONG	isEndr( char *s )
{
	return( (strnicmp(s,"Endr",4)==0) && isWhiteSpace(*(s-1)) && isWhiteSpace(s[4]) );
}

void	copyrept( void )
{
	SLONG	level=1, len, instring=0;
	char	*src=pCurrentBuffer->pBuffer;

	while( *src && level )
	{
		if( instring==0 )
		{
			if( isRept(src) )
			{
				level+=1;
				src+=4;
			}
			else if( isEndr(src) )
			{
				level-=1;
				src+=4;
			}
			else
			{
				if( *src=='\"' )
					instring=1;
				src+=1;
			}
		}
		else
		{
			if( *src=='\\' )
			{
				src+=2;
			}
			else if( *src=='\"' )
			{
				src+=1;
				instring=0;
			}
			else
			{
				src+=1;
			}
		}
	}

	len=src-pCurrentBuffer->pBuffer-4;

	src=pCurrentBuffer->pBuffer;
	ulNewMacroSize=len;

	if( (tzNewMacro=(char *)malloc(ulNewMacroSize+1))!=NULL )
	{
		ULONG i;

		tzNewMacro[ulNewMacroSize]=0;
		for( i=0; i<ulNewMacroSize; i+=1 )
		{
			if( (tzNewMacro[i]=src[i])=='\n' )
				nLineNo+=1;
		}
	}
	else
		fatalerror( "No mem for REPT block" );

	yyskipbytes( ulNewMacroSize+4 );

}

ULONG	isMacro( char *s )
{
	return( (strnicmp(s,"MACRO",4)==0) && isWhiteSpace(*(s-1)) && isWhiteSpace(s[5]) );
}

ULONG	isEndm( char *s )
{
	return( (strnicmp(s,"Endm",4)==0) && isWhiteSpace(*(s-1)) && isWhiteSpace(s[4]) );
}

void	copymacro( void )
{
	SLONG	level=1, len, instring=0;
	char	*src=pCurrentBuffer->pBuffer;

	while( *src && level )
	{
		if( instring==0 )
		{
			if( isMacro(src) )
			{
				level+=1;
				src+=4;
			}
			else if( isEndm(src) )
			{
				level-=1;
				src+=4;
			}
			else
			{
				if( *src=='\"' )
					instring=1;
				src+=1;
			}
		}
		else
		{
			if( *src=='\\' )
			{
				src+=2;
			}
			else if( *src=='\"' )
			{
				src+=1;
				instring=0;
			}
			else
			{
				src+=1;
			}
		}
	}

	len=src-pCurrentBuffer->pBuffer-4;

	src=pCurrentBuffer->pBuffer;
	ulNewMacroSize=len;

	if( (tzNewMacro=(char *)malloc(ulNewMacroSize+2))!=NULL )
	{
		ULONG i;

		tzNewMacro[ulNewMacroSize]='\n';
		tzNewMacro[ulNewMacroSize+1]=0;
		for( i=0; i<ulNewMacroSize; i+=1 )
		{
			if( (tzNewMacro[i]=src[i])=='\n' )
				nLineNo+=1;
		}
	}
	else
		fatalerror( "No mem for MACRO definition" );

	yyskipbytes( ulNewMacroSize+4 );
}

ULONG	isIf( char *s )
{
	return( (strnicmp(s,"If",2)==0) && isWhiteSpace(*(s-1)) && isWhiteSpace(s[2]) );
}

ULONG	isElse( char *s )
{
	return( (strnicmp(s,"Else",4)==0) && isWhiteSpace(*(s-1)) && isWhiteSpace(s[4]) );
}

ULONG	isEndc( char *s )
{
	return( (strnicmp(s,"Endc",4)==0) && isWhiteSpace(*(s-1)) && isWhiteSpace(s[4]) );
}

void	if_skip_to_else( void )
{
	SLONG	level=1, len, instring=0;
	char	*src=pCurrentBuffer->pBuffer;

	while( *src && level )
	{
		if( *src=='\n' )
			nLineNo+=1;

		if( instring==0 )
		{
			if( isIf(src) )
			{
				level+=1;
				src+=2;
			}
			else if( level==1 && isElse(src) )
			{
				level-=1;
				src+=4;
			}
			else if( isEndc(src) )
			{
				level-=1;
				if( level!=0 )
					src+=4;
			}
			else
			{
				if( *src=='\"' )
					instring=1;
				src+=1;
			}
		}
		else
		{
			if( *src=='\\' )
			{
				src+=2;
			}
			else if( *src=='\"' )
			{
				src+=1;
				instring=0;
			}
			else
			{
				src+=1;
			}
		}
	}

	len=src-pCurrentBuffer->pBuffer;

	yyskipbytes( len );
	yyunput( '\n' );
	nLineNo-=1;
}

void	if_skip_to_endc( void )
{
	SLONG	level=1, len, instring=0;
	char	*src=pCurrentBuffer->pBuffer;

	while( *src && level )
	{
		if( *src=='\n' )
			nLineNo+=1;

		if( instring==0 )
		{
			if( isIf(src) )
			{
				level+=1;
				src+=2;
			}
			else if( isEndc(src) )
			{
				level-=1;
				if( level!=0 )
					src+=4;
			}
			else
			{
				if( *src=='\"' )
					instring=1;
				src+=1;
			}
		}
		else
		{
			if( *src=='\\' )
			{
				src+=2;
			}
			else if( *src=='\"' )
			{
				src+=1;
				instring=0;
			}
			else
			{
				src+=1;
			}
		}
	}

	len=src-pCurrentBuffer->pBuffer;

	yyskipbytes( len );
	yyunput( '\n' );
	nLineNo-=1;
}


#ifdef	PCENGINE
ULONG	ZP( struct Expression *expr )
{
	return( (!rpn_isReloc(expr)) && (expr->nVal>0x1FFF) && (expr->nVal<0x2100) );
}

void	out_ZPByte( struct Expression *expr )
{
	if( rpn_isReloc(expr) )
	{
		rpn_CheckZP(expr,expr);
		out_RelByte(expr);
	}
	else
	{
		if( ZP(expr) )
			out_AbsByte(expr->nVal-0x2000);
		else
			fatalerror( "Value not in zeropage");
	}
}
#endif

%}

%union
{
    char tzSym[MAXSYMLEN+1];
    char tzString[256];
    struct Expression sVal;
    SLONG nConstValue;
}

%type	<sVal>	relocconst
%type	<nConstValue>	const
%type	<nConstValue>	const_3bit
%type	<sVal>	const_8bit
%type	<sVal>	const_16bit
%type	<sVal>	const_PCrel
%type	<nConstValue>	sectiontype

%type	<tzString>	string

%token	<nConstValue>		T_NUMBER
%token	<tzString>	T_STRING

%left	T_OP_LOGICNOT
%left	T_OP_LOGICOR T_OP_LOGICAND T_OP_LOGICEQU
%left	T_OP_LOGICGT T_OP_LOGICLT T_OP_LOGICGE T_OP_LOGICLE T_OP_LOGICNE
%left	T_OP_ADD T_OP_SUB
%left	T_OP_OR T_OP_XOR T_OP_AND
%left	T_OP_SHL T_OP_SHR
%left	T_OP_MUL T_OP_DIV T_OP_MOD
%left	T_OP_NOT
%left	T_OP_DEF
%left	T_OP_BANK
%left	T_OP_SIN
%left	T_OP_COS
%left	T_OP_TAN
%left	T_OP_ASIN
%left	T_OP_ACOS
%left	T_OP_ATAN
%left	T_OP_ATAN2
%left	T_OP_FDIV
%left	T_OP_FMUL

%left	T_OP_STRCMP
%left	T_OP_STRIN
%left	T_OP_STRSUB
%left	T_OP_STRLEN
%left	T_OP_STRCAT
%left	T_OP_STRUPR
%left	T_OP_STRLWR

%left	NEG     /* negation--unary minus */

%token	<tzSym> T_LABEL
%token	<tzSym> T_ID
%token	<tzSym> T_POP_EQU
%token	<tzSym> T_POP_SET
%token	<tzSym> T_POP_EQUS

%token	T_POP_INCLUDE T_POP_PRINTF T_POP_PRINTT T_POP_PRINTV T_POP_IF T_POP_ELSE T_POP_ENDC
%token	T_POP_IMPORT T_POP_EXPORT T_POP_GLOBAL
%token	T_POP_DB T_POP_DS T_POP_DW T_POP_DL
%token	T_POP_SECTION
%token	T_POP_RB
%token	T_POP_RW
%token	T_POP_RL
%token	T_POP_MACRO
%token	T_POP_ENDM
%token	T_POP_RSRESET T_POP_RSSET
%token	T_POP_INCBIN T_POP_REPT
%token	T_POP_SHIFT
%token	T_POP_ENDR
%token	T_POP_FAIL
%token	T_POP_WARN
%token	T_POP_PURGE
%token	T_POP_POPS
%token	T_POP_PUSHS
%token	T_POP_POPO
%token	T_POP_PUSHO
%token	T_POP_OPT%token	T_SECT_BSS T_SECT_VRAM T_SECT_CODE T_SECT_SAVESSCREEN T_SECT_STATVARS T_SECT_TEXTSHADOW T_SECT_PLOTSCREEN

%token	T_Z80_ADC T_Z80_ADD T_Z80_AND
%token	T_Z80_BIT
%token	T_Z80_CALL T_Z80_CCF T_Z80_CP T_Z80_CPL
%token	T_Z80_CPD T_Z80_CPDR T_Z80_CPI T_Z80_CPIR
%token	T_Z80_DAA T_Z80_DEC T_Z80_DI T_Z80_DJNZ
%token	T_Z80_EI T_Z80_EX T_Z80_EXX
%token	T_Z80_HALT
%token	T_Z80_IM T_Z80_IN T_Z80_INC
%token	T_Z80_IND T_Z80_INDR T_Z80_INI T_Z80_INIR
%token	T_Z80_JP T_Z80_JR
%token	T_Z80_LD
%token	T_Z80_LDD T_Z80_LDDR
%token	T_Z80_LDI T_Z80_LDIR
%token	T_Z80_NEG T_Z80_NOP
%token	T_Z80_OR T_Z80_OUT
%token	T_Z80_OUTD T_Z80_OTDR T_Z80_OUTI T_Z80_OTIR
%token	T_Z80_POP T_Z80_PUSH
%token	T_Z80_RES T_Z80_RET T_Z80_RETI T_Z80_RETN T_Z80_RST
%token	T_Z80_RL T_Z80_RLA T_Z80_RLC T_Z80_RLCA
%token	T_Z80_RR T_Z80_RRA T_Z80_RRC T_Z80_RRCA
%token	T_Z80_RLD T_Z80_RRD
%token	T_Z80_SBC T_Z80_SCF
%token	T_Z80_SLA T_Z80_SRA T_Z80_SRL T_Z80_SUB
%token	T_Z80_XOR

%token	T_MODE_A T_MODE_B T_MODE_C T_MODE_C_IND T_MODE_D T_MODE_E T_MODE_H T_MODE_L
%token	T_MODE_AF T_MODE_AFP
%token	T_MODE_I
%token	T_MODE_IX T_MODE_IY
%token	T_MODE_BC T_MODE_BC_IND
%token	T_MODE_DE T_MODE_DE_IND
%token	T_MODE_SP T_MODE_SP_IND
%token	T_MODE_HL T_MODE_HL_IND
%token	T_CC_NZ T_CC_Z T_CC_NC T_CC_PO T_CC_PE T_CC_P T_CC_M

%type	<nConstValue>	reg_r
%type	<nConstValue>	reg_i
%type	<nConstValue>	reg_ss
%type	<nConstValue>	reg_rr
%type	<nConstValue>	reg_tt
%type	<nConstValue>	ccode
%type	<sVal>		op_a_n
%type	<nConstValue>	op_a_r
%type	<nConstValue>	op_hl_ss
%type	<sVal>		op_mem_ind
%type	<sVal>		op_indexed
%start asmfile

%%

asmfile			:	lines lastline
;

lastline		:	/* empty */
				|	line
					{ nLineNo+=1; nTotalLines+=1; }
;

lines			:	/* empty */
				|	lines line '\n'
					{ nLineNo+=1; nTotalLines+=1; }
;

line			:	/* empty */
				|	label
				|	label cpu_command
				|	label macro
				|	label simple_pseudoop
				|	pseudoop
;

label			:	/* empty */
				|	T_LABEL			{	if( $1[0]=='.' )
											sym_AddLocalReloc($1);
										else
											sym_AddReloc($1);
									}
				|	T_LABEL ':'		{	if( $1[0]=='.' )
											sym_AddLocalReloc($1);
										else
											sym_AddReloc($1);
									}
				|	T_LABEL ':' ':'	{ sym_AddReloc($1); sym_Export($1); }
;

macro			:	T_ID
					{
						yy_set_state( LEX_STATE_MACROARGS );
					}
					macroargs
					{
						yy_set_state( LEX_STATE_NORMAL );

						if( !fstk_RunMacro($1) )
						{
							yyerror( "No such macro" );
						}
					}
;

macroargs		:	/* empty */
				|	macroarg
				|	macroarg ',' macroargs
;

macroarg 		:	T_STRING
					{ sym_AddNewMacroArg( $1 ); }
;

pseudoop		:	equ
				|	set
				|	rb
				|	rw
				|	rl
				|	equs
				|	macrodef
;

simple_pseudoop	:	include
				|	printf
				|	printt
				|	printv
				|	if
				|	else
				|	endc
				|	import
				|	export
				|	global
				|	db
				|	dw
				|	dl
				|	ds
				|	section
				|	rsreset
				|	rsset
				|	incbin
				|	rept
				|	shift
				|	fail
				|	warn
				|	purge
				|	pops
				|	pushs
				|	popo
				|	pusho
				|	opt
;

opt				:	T_POP_OPT
					{
						yy_set_state( LEX_STATE_MACROARGS );
					}
					opt_list
					{
						yy_set_state( LEX_STATE_NORMAL );
					}
;

opt_list		:	opt_list_entry
				|	opt_list_entry ',' opt_list
;

opt_list_entry	:	T_STRING
					{
						opt_Parse($1);
					}
;

popo			:	T_POP_POPO
					{ opt_Pop(); }
;

pusho			:	T_POP_PUSHO
					{ opt_Push(); }
;

pops			:	T_POP_POPS
					{ out_PopSection(); }
;

pushs			:	T_POP_PUSHS
					{ out_PushSection(); }
;

fail			:	T_POP_FAIL string
					{ fatalerror($2); }
;

warn			:	T_POP_WARN string
					{ yyerror($2); }
;

shift			:	T_POP_SHIFT
					{ sym_ShiftCurrentMacroArgs(); }
;

rept			:	T_POP_REPT const
					{
						copyrept();
						fstk_RunRept( $2 );
					}
;

macrodef		:	T_LABEL ':' T_POP_MACRO
					{
						copymacro();
						sym_AddMacro($1);
					}
;

equs			:	T_LABEL T_POP_EQUS string
					{ sym_AddString( $1, $3 ); }
;

rsset			:	T_POP_RSSET const
					{ sym_AddSet( "_RS", $2 ); }
;

rsreset			:	T_POP_RSRESET
					{ sym_AddSet( "_RS", 0 ); }
;

rl				:	T_LABEL T_POP_RL const
					{
						sym_AddEqu( $1, sym_GetConstantValue("_RS") );
						sym_AddSet( "_RS", sym_GetConstantValue("_RS")+4*$3 );
					}
;

rw				:	T_LABEL T_POP_RW const
					{
						sym_AddEqu( $1, sym_GetConstantValue("_RS") );
						sym_AddSet( "_RS", sym_GetConstantValue("_RS")+2*$3 );
					}
;

rb				:	T_LABEL T_POP_RB const
					{
						sym_AddEqu( $1, sym_GetConstantValue("_RS") );
						sym_AddSet( "_RS", sym_GetConstantValue("_RS")+$3 );
					}
;

ds				:	T_POP_DS const
					{ out_Skip( $2 ); }
;

db				:	T_POP_DB constlist_8bit
;

dw				:	T_POP_DW constlist_16bit
;

dl				:	T_POP_DL constlist_32bit
;

purge			:	T_POP_PURGE
					{
						oDontExpandStrings=1;
					}
					purge_list
					{
						oDontExpandStrings=0;
					}
;

purge_list		:	purge_list_entry
				|	purge_list_entry ',' purge_list
;

purge_list_entry	:	T_ID	{ sym_Purge($1); }
;

import			:	T_POP_IMPORT import_list
;

import_list		:	import_list_entry
				|	import_list_entry ',' import_list
;

import_list_entry	:	T_ID	{ sym_Import($1); }
;

export			:	T_POP_EXPORT export_list
;

export_list		:	export_list_entry
				|	export_list_entry ',' export_list
;

export_list_entry	:	T_ID	{ sym_Export($1); }
;

global			:	T_POP_GLOBAL global_list
;

global_list		:	global_list_entry
				|	global_list_entry ',' global_list
;

global_list_entry	:	T_ID	{ sym_Global($1); }
;

equ				:	T_LABEL T_POP_EQU const
					{ sym_AddEqu( $1, $3 ); }
;

set				:	T_LABEL T_POP_SET const
					{ sym_AddSet( $1, $3 ); }
;

include			:	T_POP_INCLUDE string
					{
						if( !fstk_RunInclude($2) )
						{
							yyerror( "File not found" );
						}
					}
;

incbin			:	T_POP_INCBIN string
					{ out_BinaryFile( $2 ); }
;

printt			:	T_POP_PRINTT string
					{
						if( nPass==1 )
							printf( "%s", $2 );
					}
;

printv			:	T_POP_PRINTV const
					{
						if( nPass==1 )
							printf( "$%lX", $2 );
					}
;

printf			:	T_POP_PRINTF const
					{
						if( nPass==1 )
							math_Print( $2 );
					}
;

if				:	T_POP_IF const
					{
						nIFDepth+=1;
						if( !$2 )
						{
							if_skip_to_else();	/* will continue parsing just after ELSE or just at ENDC keyword */
						}
					}

else			:	T_POP_ELSE
					{
						if_skip_to_endc();		/* will continue parsing just at ENDC keyword */
					}
;

endc			:	T_POP_ENDC
					{
						nIFDepth-=1;
					}
;

const_3bit		:	const
					{
						if( ($1<0) || ($1>7) )
						{
							yyerror( "Immediate value must be 3-bit" );
						}
						else
							$$=$1&0x7;
					}
;

constlist_8bit	:	constlist_8bit_entry
				|	constlist_8bit_entry ',' constlist_8bit
;

constlist_8bit_entry	:               { out_Skip( 1 ); }
						|	const_8bit	{ out_RelByte( &$1 ); }
			   			|	string		{ out_String( $1 );  }
;

constlist_16bit	:	constlist_16bit_entry
				|	constlist_16bit_entry ',' constlist_16bit
;

constlist_16bit_entry	:  				{ out_Skip( 2 ); }
						|	const_16bit	{ out_RelWord( &$1 ); }
;


constlist_32bit	:	constlist_32bit_entry
				|	constlist_32bit_entry ',' constlist_32bit
;

constlist_32bit_entry	:				{ out_Skip( 4 ); }
						|	relocconst	{ out_RelLong( &$1 ); }
;


const_PCrel		:	relocconst
					{
						$$ = $1;
						if( !rpn_isPCRelative(&$1) )
							yyerror( "Expression must be PC-relative" );
					}
;

const_8bit		:	relocconst
					{
						if( (!rpn_isReloc(&$1)) && (($1.nVal<-128) || ($1.nVal>255)) )
						{
							yyerror( "Expression must be 8-bit" );
						}
						$$=$1;
					}
;

const_16bit		:	relocconst
					{
						if( (!rpn_isReloc(&$1)) && (($1.nVal<-32768) || ($1.nVal>65535)) )
						{
							yyerror( "Expression must be 16-bit" );
						}
						$$=$1
					}
;


relocconst		:	T_ID
						{ rpn_Symbol(&$$,$1);	$$.nVal = sym_GetValue($1); }
				|	T_NUMBER
						{ rpn_Number(&$$,$1);	$$.nVal = $1; }
				|	string
						{ ULONG r; r=str2int($1); rpn_Number(&$$,r); $$.nVal=r; }
				|	T_OP_LOGICNOT relocconst %prec NEG
						{ rpn_LOGNOT(&$$,&$2); }
				|	relocconst T_OP_LOGICOR relocconst
						{ rpn_LOGOR(&$$,&$1,&$3); }
				|	relocconst T_OP_LOGICAND relocconst
						{ rpn_LOGAND(&$$,&$1,&$3); }
				|	relocconst T_OP_LOGICEQU relocconst
						{ rpn_LOGEQU(&$$,&$1,&$3); }
				|	relocconst T_OP_LOGICGT relocconst
						{ rpn_LOGGT(&$$,&$1,&$3); }
				|	relocconst T_OP_LOGICLT relocconst
						{ rpn_LOGLT(&$$,&$1,&$3); }
				|	relocconst T_OP_LOGICGE relocconst
					 	{ rpn_LOGGE(&$$,&$1,&$3); }
				|	relocconst T_OP_LOGICLE relocconst
					 	{ rpn_LOGLE(&$$,&$1,&$3); }
				|	relocconst T_OP_LOGICNE relocconst
					 	{ rpn_LOGNE(&$$,&$1,&$3); }
				|	relocconst T_OP_ADD relocconst
						{ rpn_ADD(&$$,&$1,&$3); }
				|	relocconst T_OP_SUB relocconst
						{ rpn_SUB(&$$,&$1,&$3); }
				|	relocconst T_OP_XOR relocconst
						{ rpn_XOR(&$$,&$1,&$3); }
				|	relocconst T_OP_OR relocconst
						{ rpn_OR(&$$,&$1,&$3); }
				|	relocconst T_OP_AND relocconst
						{ rpn_AND(&$$,&$1,&$3); }
				|	relocconst T_OP_SHL relocconst
						{ rpn_SHL(&$$,&$1,&$3); }
				|	relocconst T_OP_SHR relocconst
						{ rpn_SHR(&$$,&$1,&$3); }
				|	relocconst T_OP_MUL relocconst
						{ rpn_MUL(&$$,&$1,&$3); }
				|	relocconst T_OP_DIV relocconst
						{ rpn_DIV(&$$,&$1,&$3); }
				|	relocconst T_OP_MOD relocconst
						{ rpn_MOD(&$$,&$1,&$3); }
				|	T_OP_ADD relocconst %prec NEG
						{ $$ = $2; }
				|	T_OP_SUB relocconst %prec NEG
						{ rpn_UNNEG(&$$,&$2); }
				|	T_OP_NOT relocconst %prec NEG
						{ rpn_UNNOT(&$$,&$2); }
				|	T_OP_BANK '(' T_ID ')'
						{ rpn_Bank(&$$,$3); $$.nVal = 0; }
				|	T_OP_DEF '(' T_ID ')'
						{ rpn_Number(&$$,sym_isConstDefined($3)); }
				|	T_OP_FDIV '(' const ',' const ')'			{ rpn_Number(&$$,math_Div($3,$5)); }
				|	T_OP_FMUL '(' const ',' const ')'			{ rpn_Number(&$$,math_Mul($3,$5)); }
				|	T_OP_SIN '(' const ')'					{ rpn_Number(&$$,math_Sin($3)); } 
				|	T_OP_COS '(' const ')'					{ rpn_Number(&$$,math_Cos($3)); } 
				|	T_OP_TAN '(' const ')'					{ rpn_Number(&$$,math_Tan($3)); } 
				|	T_OP_ASIN '(' const ')'					{ rpn_Number(&$$,math_ASin($3)); }
				|	T_OP_ACOS '(' const ')'					{ rpn_Number(&$$,math_ACos($3)); }
				|	T_OP_ATAN '(' const ')'					{ rpn_Number(&$$,math_ATan($3)); }
				|	T_OP_ATAN2 '(' const ',' const ')'			{ rpn_Number(&$$,math_ATan2($3,$5)); }
				|	T_OP_STRCMP '(' string ',' string ')'		{ rpn_Number(&$$,strcmp($3,$5)); }
				|	T_OP_STRIN '(' string ',' string ')'
					{
						char	*p;
  						if( (p=strstr($3,$5))!=NULL )
						{
							rpn_Number(&$$,p-$3+1);
						}
						else
						{
							rpn_Number(&$$,0);
						}
					}
				|	T_OP_STRLEN '(' string ')'		{ rpn_Number(&$$,strlen($3)); }
				|	'(' relocconst ')'
						{ $$ = $2; }
;

const			:	T_ID							{ $$ = sym_GetConstantValue($1); }
				|	T_NUMBER 						{ $$ = $1; }
				|	string						{ $$ = str2int($1); }
				|	T_OP_LOGICNOT const %prec NEG	{ $$ = !$2; }
				|	const T_OP_LOGICOR const		{ $$ = $1 || $3; }
				|	const T_OP_LOGICAND const		{ $$ = $1 && $3; }
				|	const T_OP_LOGICEQU const		{ $$ = $1 == $3; }
				|	const T_OP_LOGICGT const 		{ $$ = $1 > $3; }
				|	const T_OP_LOGICLT const		{ $$ = $1 < $3; }
				|	const T_OP_LOGICGE const 		{ $$ = $1 >= $3; }
				|	const T_OP_LOGICLE const 		{ $$ = $1 <= $3; }
				|	const T_OP_LOGICNE const 		{ $$ = $1 != $3; }
				|	const T_OP_ADD const			{ $$ = $1 + $3; }
				|	const T_OP_SUB const			{ $$ = $1 - $3; }
				|	T_ID  T_OP_SUB T_ID				{ $$ = sym_GetDefinedValue($1) - sym_GetDefinedValue($3); }
				|	const T_OP_XOR const			{ $$ = $1 ^ $3; }
				|	const T_OP_OR const				{ $$ = $1 | $3; }
				|	const T_OP_AND const			{ $$ = $1 & $3; }
				|	const T_OP_SHL const			{ $$ = $1 << $3; }
				|	const T_OP_SHR const			{ $$ = $1 >> $3; }
				|	const T_OP_MUL const			{ $$ = $1 * $3; }
				|	const T_OP_DIV const			{ $$ = $1 / $3; }
				|	const T_OP_MOD const			{ $$ = $1 % $3; }
				|	T_OP_ADD const %prec NEG		{ $$ = +$2; }
				|	T_OP_SUB const %prec NEG		{ $$ = -$2; }
				|	T_OP_NOT const %prec NEG		{ $$ = 0xFFFFFFFF^$2; }
				|	T_OP_FDIV '(' const ',' const ')'			{ $$ = math_Div($3,$5); }
				|	T_OP_FMUL '(' const ',' const ')'			{ $$ = math_Mul($3,$5); }
				|	T_OP_SIN '(' const ')'			{ $$ = math_Sin($3); }
				|	T_OP_COS '(' const ')'			{ $$ = math_Cos($3); }
				|	T_OP_TAN '(' const ')'			{ $$ = math_Tan($3); }
				|	T_OP_ASIN '(' const ')'			{ $$ = math_ASin($3); }
				|	T_OP_ACOS '(' const ')'			{ $$ = math_ACos($3); }
				|	T_OP_ATAN '(' const ')'			{ $$ = math_ATan($3); }
				|	T_OP_ATAN2 '(' const ',' const ')'	{ $$ = math_ATan2($3,$5); }
				|	T_OP_DEF '(' T_ID ')'				{ $$ = sym_isConstDefined($3); }
				|	T_OP_STRCMP '(' string ',' string ')'	{ $$ = strcmp( $3, $5 ); }
				|	T_OP_STRIN '(' string ',' string ')'
					{
						char	*p;
  						if( (p=strstr($3,$5))!=NULL )
						{
							$$ = p-$3+1;
						}
						else
						{
							$$ = 0;
						}
					}
				|	T_OP_STRLEN '(' string ')'		{ $$ = strlen($3); }
				|	'(' const ')'					{ $$ = $2; }
;

string			:	T_STRING
					{ strcpy($$,$1); }
				|	T_OP_STRSUB '(' string ',' const ',' const ')'
					{ strncpy($$,$3+$5-1,$7); $$[$7]=0; }
				|	T_OP_STRCAT '(' string ',' string ')'
					{ strcpy($$,$3); strcat($$,$5); }
				|	T_OP_STRUPR '(' string ')'
					{ strcpy($$,$3); strupr($$); }
				|	T_OP_STRLWR '(' string ')'
					{ strcpy($$,$3); strlwr($$); }
;section			:	T_POP_SECTION string ',' sectiontype
					{ out_NewSection($2,$4); }
		|	T_POP_SECTION string ',' sectiontype '[' const ']'
					{
						if( $6>=0 && $6<0x10000 )
							out_NewAbsSection($2,$4,$6,-1);
						else
							yyerror( "Address must be 16-bit" );
					}
		|	T_POP_SECTION string ',' sectiontype ',' T_OP_BANK '[' const ']'
					{
						if( $4==SECT_CODE )
						{
							if( $8>=1 && $8<=255 )
								out_NewAbsSection($2,$4,-1,$8);
							else
								yyerror( "BANK value out of range" );
						}
						else
							yyerror( "BANK only allowed for CODE/DATA" );
					}
		|	T_POP_SECTION string ',' sectiontype '[' const ']' ',' T_OP_BANK '[' const ']'
					{
						if( $4==SECT_CODE )
						{
							if( $6>=0 && $6<0x10000 )
							{
								if( $11>=1 && $11<=255 )
									out_NewAbsSection($2,$4,$6,$11);
								else
									yyerror( "BANK value out of range" );
							}
							else
								yyerror( "Address must be 16-bit" );
						}
						else
							yyerror( "BANK only allowed for CODE/DATA" );
					}
;

sectiontype		:	T_SECT_BSS	{ $$=SECT_BSS; }
			|	T_SECT_VRAM		{ $$=SECT_VRAM; 	}
			|	T_SECT_CODE		{ $$=SECT_CODE; 	}
			|	T_SECT_SAVESSCREEN	{ $$=SECT_SAVESCREEN;	}
			|	T_SECT_STATVARS		{ $$=SECT_STATVARS;	}
			|	T_SECT_TEXTSHADOW	{ $$=SECT_TEXTSHADOW;	}
			|	T_SECT_PLOTSCREEN	{ $$=SECT_PLOTSCREEN;	}
;


cpu_command		:	z80_adc
			|	z80_add
			|	z80_and
			|	z80_bit
			|	z80_call
			|	z80_ccf
			|	z80_cp
			|	z80_cpd
			|	z80_cpdr
			|	z80_cpi
			|	z80_cpir
			|	z80_cpl
			|	z80_daa
			|	z80_dec
			|	z80_di
			|	z80_djnz
			|	z80_ei
			|	z80_ex
			|	z80_exx
			|	z80_halt
			|	z80_im
			|	z80_in
			|	z80_ind
			|	z80_indr
			|	z80_ini
			|	z80_inir
			|	z80_inc
			|	z80_jp
			|	z80_jr
			|	z80_ld
			|	z80_ldd
			|	z80_lddr
			|	z80_ldi
			|	z80_ldir
			|	z80_neg
			|	z80_nop
			|	z80_or
			|	z80_out
			|	z80_outd
			|	z80_otdr
			|	z80_outi
			|	z80_otir
			|	z80_pop
			|	z80_push
			|	z80_res
			|	z80_ret
			|	z80_reti
			|	z80_retn
			|	z80_rl
			|	z80_rla
			|	z80_rlc
			|	z80_rlca
			|	z80_rld
			|	z80_rr
			|	z80_rra
			|	z80_rrc
			|	z80_rrca
			|	z80_rrd
			|	z80_rst
			|	z80_sbc
			|	z80_scf
			|	z80_set
			|	z80_sla
			|	z80_sra
			|	z80_srl
			|	z80_sub
			|	z80_xor
;

z80_adc			:	T_Z80_ADC op_a_n
					{ out_AbsByte(0xCE); out_RelByte(&$2); }
			|	T_Z80_ADC op_a_r
					{ out_AbsByte(0x88|$2); }
			|	T_Z80_ADC T_MODE_HL comma reg_ss
					{ out_AbsByte(0xED); out_AbsByte(0x4A | ($4 << 4)); }
			|	T_Z80_ADC T_MODE_A comma op_indexed
					{ out_AbsByte($4.nUserData); out_AbsByte(0x8E); out_RelByte(&$4); }
;

z80_add			:	T_Z80_ADD op_a_n	{ out_AbsByte(0xC6); out_RelByte(&$2); }
			|	T_Z80_ADD op_a_r	{ out_AbsByte(0x80|$2); }
			|	T_Z80_ADD op_hl_ss	{ out_AbsByte(0x09|($2<<4)); }
			|	T_Z80_ADD T_MODE_A comma op_indexed
					{ out_AbsByte($4.nUserData); out_AbsByte(0x86); out_RelByte(&$4); }
			|	T_Z80_ADD reg_i comma reg_ss
					{
						if($4 == REG_HL)
							yyerror("Source must not be HL!");
						else
							{ out_AbsByte($2); out_AbsByte(0x09 | ($4 << 4)); }
					}
			|	T_Z80_ADD reg_i comma reg_i
					{
						if($2 != $4)
							yyerror("Invalid operands for add!");
						else
							{ out_AbsByte($2); out_AbsByte(0x29); }
					}

;

z80_and			:	T_Z80_AND op_a_n	{ out_AbsByte(0xE6); out_RelByte(&$2); }
			|	T_Z80_AND op_a_r	{ out_AbsByte(0xA0|$2); }
			|	T_Z80_AND op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xA6); out_RelByte(&$2); }
;

z80_bit			:	T_Z80_BIT const_3bit comma reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x40|($2<<3)|$4); }
			|	T_Z80_BIT const_3bit comma op_indexed
					{ out_AbsByte($4.nUserData); out_AbsByte(0xCB); out_RelByte(&$4); out_AbsByte(0x46 | ($2 << 3)); }
;

z80_call		:	T_Z80_CALL const_16bit
					{ out_AbsByte(0xCD); out_RelWord(&$2); }
			|	T_Z80_CALL ccode comma const_16bit
					{ out_AbsByte(0xC4|($2<<3)); out_RelWord(&$4); }
;

z80_ccf			:	T_Z80_CCF
					{ out_AbsByte(0x3F); }
;

z80_cp			:	T_Z80_CP op_a_n	{ out_AbsByte(0xFE); out_RelByte(&$2); }
			|	T_Z80_CP op_a_r	{ out_AbsByte(0xB8|$2); }
			|	T_Z80_CP op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xBE); out_RelByte(&$2); }
;

z80_cpd			:	T_Z80_CPD
					{ out_AbsByte(0xED); out_AbsByte(0xA9); }
;

z80_cpdr		:	T_Z80_CPDR
					{ out_AbsByte(0xED); out_AbsByte(0xB9); }
;

z80_cpi			:	T_Z80_CPI
					{ out_AbsByte(0xED); out_AbsByte(0xA1); }
;

z80_cpir		:	T_Z80_CPIR
					{ out_AbsByte(0xED); out_AbsByte(0xB1); }
;

z80_cpl			:	T_Z80_CPL { out_AbsByte(0x2F); }
;

z80_daa			:	T_Z80_DAA { out_AbsByte(0x27); }
;

z80_dec			:	T_Z80_DEC reg_r
					{ out_AbsByte(0x05|($2<<3)); }
			|	T_Z80_DEC reg_ss
					{ out_AbsByte(0x0B|($2<<4)); }
			|	T_Z80_DEC reg_i
					{ out_AbsByte($2); out_AbsByte(0x2B); }						
			|	T_Z80_DEC op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0x35); out_RelByte(&$2); }
;

z80_di			:	T_Z80_DI
					{ out_AbsByte(0xF3); }
;

z80_djnz			:	T_Z80_DJNZ const_PCrel
					{ out_AbsByte(0x10); out_PCRelByte(&$2); }
;

z80_ei			:	T_Z80_EI
					{ out_AbsByte(0xFB); }
;

z80_ex			:	T_Z80_EX T_MODE_HL comma T_MODE_SP_IND
					{ out_AbsByte(0xE3); }
			|	T_Z80_EX T_MODE_SP_IND comma T_MODE_HL
					{ out_AbsByte(0xE3); }
			|	T_Z80_EX T_MODE_AF comma T_MODE_AFP
					{ out_AbsByte(0x08); }
			|	T_Z80_EX T_MODE_AFP comma T_MODE_AF
					{ out_AbsByte(0x08); }
			|	T_Z80_EX T_MODE_DE comma T_MODE_HL
					{ out_AbsByte(0xEB); }
			|	T_Z80_EX T_MODE_HL comma T_MODE_DE
					{ out_AbsByte(0xEB); }
			|	T_Z80_EX T_MODE_SP_IND comma T_MODE_IX
					{ out_AbsByte(0xDD); out_AbsByte(0xE3); }
			|	T_Z80_EX T_MODE_IX comma T_MODE_SP_IND
					{ out_AbsByte(0xDD); out_AbsByte(0xE3); }
			|	T_Z80_EX T_MODE_SP_IND comma T_MODE_IY
					{ out_AbsByte(0xFD); out_AbsByte(0xE3); }
			|	T_Z80_EX T_MODE_IY comma T_MODE_SP_IND
					{ out_AbsByte(0xFD); out_AbsByte(0xE3); }
;

z80_exx			:	T_Z80_EXX
					{ out_AbsByte(0xD9); }
;

z80_halt			:	T_Z80_HALT
					{ out_AbsByte(0x76); out_AbsByte(0x00); }
;

z80_im			:	T_Z80_IM const_8bit
					{
					if( rpn_isReloc(&$2) )
						yyerror( "Value of x for 'IM x' must be constant!" );
					else if( $2.nVal == 0 )
						{ out_AbsByte(0xED); out_AbsByte(0x46); }
					else if( $2.nVal == 1 )
						{ out_AbsByte(0xED); out_AbsByte(0x56); }
					else if( $2.nVal == 2 )
						{ out_AbsByte(0xED); out_AbsByte(0x5E); }
					else
						yyerror( "Invalid value for x in 'IM x'!");
					}

;

z80_in			:	T_Z80_IN reg_r comma '[' const_8bit ']'
					{
						if( $2==REG_A )
							{ out_AbsByte(0xDB); out_RelByte(&$5); }
						else
						{
							yyerror( "Destination operand must be A" );
						}
					}
			|	T_Z80_IN reg_r comma T_MODE_C_IND
					{ out_AbsByte(0xED); out_AbsByte(0x40|($2<<3)); }
;


z80_inc			:	T_Z80_INC reg_r
					{ out_AbsByte(0x04|($2<<3)); }
			|	T_Z80_INC reg_ss
					{ out_AbsByte(0x03|($2<<4)); }
			|	T_Z80_INC reg_i
					{ out_AbsByte($2); out_AbsByte(0x23); }
			|	T_Z80_INC op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0x34); out_RelByte(&$2); }
;

z80_ind			:	T_Z80_IND
					{ out_AbsByte(0xED); out_AbsByte(0xAA); }
;

z80_indr		:	T_Z80_INDR
					{ out_AbsByte(0xED); out_AbsByte(0xBA); }
;

z80_ini			:	T_Z80_INI
					{ out_AbsByte(0xED); out_AbsByte(0xA2); }
;

z80_inir		:	T_Z80_INIR
					{ out_AbsByte(0xED); out_AbsByte(0xB2); }
;

z80_jp			:	T_Z80_JP const_16bit
					{ out_AbsByte(0xC3); out_RelWord(&$2); }
			|	T_Z80_JP ccode comma const_16bit
					{ out_AbsByte(0xC2|($2<<3)); out_RelWord(&$4); }
			|	T_Z80_JP T_MODE_HL_IND
					{ out_AbsByte(0xE9); }
			|	T_Z80_JP '[' reg_i ']'
					{ out_AbsByte($3); out_AbsByte(0xE9); }
;

z80_jr			:	T_Z80_JR const_PCrel
					{ out_AbsByte(0x18); out_PCRelByte(&$2); }
			|	T_Z80_JR ccode comma const_PCrel
					{ out_AbsByte(0x20|($2<<3)); out_PCRelByte(&$4); }
;

z80_ldd			:	T_Z80_LDD
					{ out_AbsByte(0xED); out_AbsByte(0xA8); }
;

z80_lddr		:	T_Z80_LDDR
					{ out_AbsByte(0xED); out_AbsByte(0xB8); }
;

z80_ldi			:	T_Z80_LDI
					{ out_AbsByte(0xED); out_AbsByte(0xA0); }
;

z80_ldir		:	T_Z80_LDIR
					{ out_AbsByte(0xED); out_AbsByte(0xB0); }
;

z80_ld			:	z80_ld_mem
			|	z80_ld_rr
			|	z80_ld_hl
			|	z80_ld_sp
			|	z80_ld_r
			|	z80_ld_i
			|	z80_ld_a
			|	z80_ld_ss
			|	z80_ld_indexed
;

z80_ld_hl		:	T_Z80_LD T_MODE_HL comma const_16bit
					{ out_AbsByte(0x01|(REG_HL<<4)); out_RelWord(&$4) }
			|	T_Z80_LD T_MODE_HL comma op_mem_ind
					{ out_AbsByte(0xED); out_AbsByte(0x4B | (REG_HL << 4)); out_RelWord(&$4) }
;
z80_ld_sp		:	T_Z80_LD T_MODE_SP comma T_MODE_HL
					{ out_AbsByte(0xF9); }
			|	T_Z80_LD T_MODE_SP comma const_16bit
					{ out_AbsByte(0x01|(REG_SP<<4)); out_RelWord(&$4) }
			|	T_Z80_LD T_MODE_SP comma op_mem_ind
					{ out_AbsByte(0xED); out_AbsByte(0x4B | (REG_SP << 4)); out_RelWord(&$4) }
			|	T_Z80_LD T_MODE_SP comma reg_i
					{ out_AbsByte($4); out_AbsByte(0xF9); }
;

z80_ld_mem		:	T_Z80_LD op_mem_ind comma reg_ss
					{ out_AbsByte(0x0ED); out_AbsByte(0x43 | ($4 << 4)); out_RelWord(&$2); }
			|	T_Z80_LD op_mem_ind comma T_MODE_A
					{	out_AbsByte(0x32); out_RelWord(&$2); }
			|	T_Z80_LD op_mem_ind comma reg_i
					{ out_AbsByte($4); out_AbsByte(0x22); out_RelWord(&$2); }
;

z80_ld_rr		:	T_Z80_LD reg_rr comma T_MODE_A
					{ out_AbsByte(0x02|($2<<4)); }
;

z80_ld_r		:	T_Z80_LD reg_r comma const_8bit
					{ out_AbsByte(0x06|($2<<3)); out_RelByte(&$4); }
			|	T_Z80_LD reg_r comma reg_r
					{
						if( ($2==REG_HL_IND) && ($4==REG_HL_IND) )
						{
							yyerror( "LD [HL],[HL] not allowed" );
						}
						else
							out_AbsByte(0x40|($2<<3)|$4);
					}
			|	T_Z80_LD reg_r comma op_indexed
					{
						if($2 == REG_HL_IND)
							yyerror("Can not load into [HL] with indexed adressing");
						else
							{ out_AbsByte($4.nUserData); out_AbsByte(0x46 | ($2 << 3)); out_RelByte(&$4); }
					}
			|	T_Z80_LD reg_i comma op_mem_ind
					{ out_AbsByte($2); out_AbsByte(0x2A); out_RelWord(&$4); }
;

z80_ld_i		:	T_Z80_LD T_MODE_I comma T_MODE_A
					{ out_AbsByte(0xED); out_AbsByte(0x47); }
;

z80_ld_a		:	T_Z80_LD reg_r comma reg_rr
					{
						if( $2==REG_A )
							out_AbsByte(0x0A|($4<<4));
						else
						{
							yyerror( "Destination operand must be A" );
						}
					}
			|	T_Z80_LD reg_r comma op_mem_ind
					{
						if( $2==REG_A )
						{
							out_AbsByte(0x3A);
							out_RelWord(&$4);
						}
						else
						{
							yyerror( "Destination operand must be A" );
						}
					}
			|	T_Z80_LD reg_r comma T_MODE_I
					{
						if( $2==REG_A )
							{ out_AbsByte(0xED); out_AbsByte(0x57); }
						else
							yyerror( "Destination operand must be A" );
					}

;

z80_ld_ss		:	T_Z80_LD reg_ss comma const_16bit
					{ out_AbsByte(0x01|($2<<4)); out_RelWord(&$4) }
			|	T_Z80_LD reg_ss comma op_mem_ind
					{ out_AbsByte(0xED); out_AbsByte(0x4B | ($2 << 4)); out_RelWord(&$4) }
			|	T_Z80_LD reg_i comma const_16bit
					{ out_AbsByte($2); out_AbsByte(0x21); out_RelWord(&$4); }

;

z80_ld_indexed		:	T_Z80_LD op_indexed comma const_8bit
					{ out_AbsByte($2.nUserData); out_AbsByte(0x36); out_RelByte(&$2); out_RelByte(&$4); }
			|	T_Z80_LD op_indexed comma reg_r
					{
						if($4 == REG_HL_IND)
							yyerror("Can not load from [HL] with indexed adressing");
						else
							{ out_AbsByte($2.nUserData); out_AbsByte(0x70 | $4); out_RelByte(&$2); }
					}

z80_neg			:	T_Z80_NEG
					{ out_AbsByte(0xED); out_AbsByte(0x44); }

z80_nop			:	T_Z80_NOP
					{ out_AbsByte(0x00); }
;

z80_or			:	T_Z80_OR op_a_n
					{ out_AbsByte(0xF6); out_RelByte(&$2); }
			|	T_Z80_OR op_a_r
					{ out_AbsByte(0xB0|$2); }
			|	T_Z80_OR op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xB6); out_RelByte(&$2); }
;

z80_out			:	T_Z80_OUT '[' const_8bit ']' comma reg_r
					{
						if( $6==REG_A )
							{ out_AbsByte(0xD3); out_RelByte(&$3); }
						else
						{
							yyerror( "Source operand must be A" );
						}
					}
			|	T_Z80_OUT T_MODE_C_IND comma reg_r
					{ out_AbsByte(0xED); out_AbsByte(0x41|($4 << 3)); }
;

z80_outd		:	T_Z80_OUTD
					{ out_AbsByte(0xED); out_AbsByte(0xAB); }
;

z80_otdr		:	T_Z80_OTDR
					{ out_AbsByte(0xED); out_AbsByte(0xBB); }
;

z80_outi		:	T_Z80_OUTI
					{ out_AbsByte(0xED); out_AbsByte(0xA3); }
;

z80_otir		:	T_Z80_OTIR
					{ out_AbsByte(0xED); out_AbsByte(0xB3); }

z80_pop			:	T_Z80_POP reg_tt
					{ out_AbsByte(0xC1|($2<<4)); }
			|	T_Z80_POP reg_i
					{ out_AbsByte($2); out_AbsByte(0xE1); }
;

z80_push		:	T_Z80_PUSH reg_tt
					{ out_AbsByte(0xC5|($2<<4)); }
			|	T_Z80_PUSH reg_i
					{ out_AbsByte($2); out_AbsByte(0xE5); }					
;

z80_res			:	T_Z80_RES const_3bit comma reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x80|($2<<3)|$4); }
			|	T_Z80_RES const_3bit comma op_indexed
					{ out_AbsByte($4.nUserData); out_AbsByte(0xCB); out_RelByte(&$4); out_AbsByte(0x86 | ($2 << 3)); }
;

z80_ret			:	T_Z80_RET
					{ out_AbsByte(0xC9); }
			|	T_Z80_RET ccode
					{ out_AbsByte(0xC0|($2<<3)); }
;

z80_reti		:	T_Z80_RETI
					{ out_AbsByte(0xED); out_AbsByte(0x4D);}
;

z80_retn		:	T_Z80_RETN
					{ out_AbsByte(0xED); out_AbsByte(0x45);}
;

z80_rl			:	T_Z80_RL reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x10|$2); }
			|	T_Z80_RL op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xCB); out_RelByte(&$2); out_AbsByte(0x16); }
;

z80_rla			:	T_Z80_RLA
					{ out_AbsByte(0x17); }
;

z80_rlc			:	T_Z80_RLC reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x00|$2); }
			|	T_Z80_RLC op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xCB); out_RelByte(&$2); out_AbsByte(0x06); }
;

z80_rlca 		:	T_Z80_RLCA
					{ out_AbsByte(0x07); }
;

z80_rld			:	T_Z80_RLD
					{ out_AbsByte(0xED); out_AbsByte(0x6F); }
;

z80_rr			:	T_Z80_RR reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x18|$2); }
			|	T_Z80_RR op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xCB); out_RelByte(&$2); out_AbsByte(0x1E); }
;

z80_rra			:	T_Z80_RRA
					{ out_AbsByte(0x1F); }
;

z80_rrc			:	T_Z80_RRC reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x08|$2); }
			
			|	T_Z80_RRC op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xCB); out_RelByte(&$2); out_AbsByte(0x0E); }
;

z80_rrca 		:	T_Z80_RRCA
					{ out_AbsByte(0x0F); }
;

z80_rrd			:	T_Z80_RRD
					{ out_AbsByte(0xED); out_AbsByte(0x67); }
;

z80_rst 		:	T_Z80_RST const_8bit
					{
						if( rpn_isReloc(&$2) )
						{
							yyerror( "Address for RST must be absolute" );
						}
						else if( ($2.nVal&0x38)!=$2.nVal )
						{
							yyerror( "Invalid address for RST" );
						}
						else
							out_AbsByte(0xC7|$2.nVal);
					}
;

z80_sbc			:	T_Z80_SBC op_a_n
					{ out_AbsByte(0xDE); out_RelByte(&$2); }
			|	T_Z80_SBC op_a_r
					{ out_AbsByte(0x98|$2); }
			|	T_Z80_SBC T_MODE_HL comma reg_ss
					{ out_AbsByte(0xED); out_AbsByte(0x42 | ($4 << 4)); }
			|	T_Z80_SBC T_MODE_A comma op_indexed
					{ out_AbsByte($4.nUserData); out_AbsByte(0x9E); out_RelByte(&$4); }
;

z80_scf			:	T_Z80_SCF
					{ out_AbsByte(0x37); }
;

z80_set			:	T_POP_SET const_3bit comma reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0xC0|($2<<3)|$4); }
			|	T_POP_SET const_3bit comma op_indexed
					{ out_AbsByte($4.nUserData); out_AbsByte(0xCB); out_RelByte(&$4); out_AbsByte(0xC6 | ($2 << 3)); }
;

z80_sla			:	T_Z80_SLA reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x20|$2); }
			|	T_Z80_SLA op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xCB); out_RelByte(&$2); out_AbsByte(0x26); }
;

z80_sra			:	T_Z80_SRA reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x28|$2); }
			|	T_Z80_SRA op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xCB); out_RelByte(&$2); out_AbsByte(0x2E); }
;

z80_srl			:	T_Z80_SRL reg_r
					{ out_AbsByte(0xCB); out_AbsByte(0x38|$2); }
;

z80_sub			:	T_Z80_SUB op_a_n	{ out_AbsByte(0xD6); out_RelByte(&$2); }
			|	T_Z80_SUB op_a_r	{ out_AbsByte(0x90|$2); }
			|	T_Z80_SUB T_MODE_A comma op_indexed
					{ out_AbsByte($4.nUserData); out_AbsByte(0x96); out_RelByte(&$4); }
;

z80_xor			:	T_Z80_XOR op_a_n	{ out_AbsByte(0xEE); out_RelByte(&$2); }
			|	T_Z80_XOR op_a_r	{ out_AbsByte(0xA8|$2); }
			|	T_Z80_XOR op_indexed
					{ out_AbsByte($2.nUserData); out_AbsByte(0xAE); out_RelByte(&$2); }
;

op_mem_ind		:	'[' const_16bit ']'	{ $$ = $2 }
;

op_indexed		:	'[' T_MODE_IX const_8bit ']' { $$ = $3; $$.nUserData = 0xDD; }
			|	'[' T_MODE_IY const_8bit ']' { $$ = $3; $$.nUserData = 0xFD; }
;

op_hl_ss 		:	reg_ss					{ $$ = $1 }
			|	T_MODE_HL comma reg_ss	{ $$ = $3 }
;

op_a_r			:	reg_r				{ $$ = $1 }
			|	T_MODE_A comma reg_r	{ $$ = $3 }
;

op_a_n			:	const_8bit				{ $$ = $1 }
			|	T_MODE_A comma const_8bit	{ $$ = $3 }
;

comma			:	','
;

ccode			:	T_CC_NZ		{ $$ = CC_NZ }
			|	T_CC_Z		{ $$ = CC_Z  }
			|	T_CC_NC		{ $$ = CC_NC }
			|	T_MODE_C	{ $$ = CC_C  }
			|	T_CC_PO		{ $$ = CC_PO }
			|	T_CC_PE		{ $$ = CC_PE }
			|	T_CC_P		{ $$ = CC_P  }
			|	T_CC_M		{ $$ = CC_M }
;

reg_r			:	T_MODE_B		{ $$ = REG_B }
			|	T_MODE_C		{ $$ = REG_C }
			|	T_MODE_D		{ $$ = REG_D }
			|	T_MODE_E		{ $$ = REG_E }
			|	T_MODE_H		{ $$ = REG_H }
			|	T_MODE_L		{ $$ = REG_L }
			|	T_MODE_HL_IND		{ $$ = REG_HL_IND }
			|	T_MODE_A		{ $$ = REG_A }
;

reg_i			:	T_MODE_IX		{ $$ = 0xDD }
			|	T_MODE_IY		{ $$ = 0xFD }
;

reg_tt			:	T_MODE_BC		{ $$ = REG_BC }
			|	T_MODE_DE		{ $$ = REG_DE }
			|	T_MODE_HL		{ $$ = REG_HL }
			|	T_MODE_AF		{ $$ = REG_AF }
;

reg_ss			:	T_MODE_BC		{ $$ = REG_BC }
			|	T_MODE_DE		{ $$ = REG_DE }
			|	T_MODE_HL		{ $$ = REG_HL }
			|	T_MODE_SP		{ $$ = REG_SP }
;

reg_rr			:	T_MODE_BC_IND		{ $$ = REG_BC_IND }
			|	T_MODE_DE_IND		{ $$ = REG_DE_IND }
;

%%