#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#include	"object.h"
#include	"output.h"
#include	"assign.h"
#include	"patch.h"
#include	"asmotor.h"
#include	"mylink.h"
#include	"mapfile.h"
#include	"main.h"
#include	"library.h"

//	Quick and dirty...but it works
#ifdef __GNUC__
#define	strcmpi	strcasecmp
#endif

enum	eBlockType
{
	BLOCK_COMMENT,
	BLOCK_OBJECTS,
	BLOCK_LIBRARIES,
	BLOCK_OUTPUT
};

SLONG	options=0;
SLONG	fillchar=-1;
enum	eOutputType	outputtype=OUTPUT_GBROM;
char	temptext[1024];
char	smartlinkstartsymbol[256];
BBOOL	usingVerboseMode = 0;

/*
 * Print out an errormessage
 *
 */

void	fatalerror( char *s )
{
	printf( "*ERROR* : %s\n", s );
	exit( 5 );
}

/*
 * ASJ: Print out a warning
 *
 */

void warning( char *s )
{
	printf( "*WARNING* : %s\n", s );
}

/*
 * ASJ: Print out an info string.
 *		But only in verbose mode.
 *
 */
void verbose( char *s)
{
	if(usingVerboseMode)
		printf( "*INFO* : %s\n", s );
}

/*
 * Print the usagescreen
 *
 */

static char usage[]="\
Usage: xlink [options] -l <linkfile> [options]\
Options:\n\
  -h              This text\n\
  -?              This text\n\
  -m <mapfile>    Write a mapfile\n\
  -n <symfile>    Write a NO$GMB compatible symfile\n\
  -z <hx>         Set the byte value (hex format) used for uninitialised\n\
                   data (default is ? for random)\n\
  -s <symbol>     Perform smart linking starting with <symbol>\n\
  -t <target>     Output target\n\
     gb           Gameboy ROM image(default)\n\
     gb-s         Gameboy small mode (32kB)\n\
     p2           Psion2 reloc module\n\
     ti83         TI-83 Program\n\
	 ti83+        TI-83+ Program\n\
  -u              Do not protect the program (TI-8x only)\n\
  -U              Do not sqish the program (TI-8x only)\n\
  -v              Verbose mode\n\
";

void	PrintUsage( void )
{
	printf ("xLink v" LINK_VERSION " (part of ASMotor " ASMOTOR_VERSION ")\n\n");

	printf(usage);

	exit( 0 );
}


/*
 * Parse the linkfile and load all the objectfiles
 *
 */

void	ProcessLinkfile( char *tzLinkfile )
{
	FILE	*pLinkfile;
	enum	eBlockType	CurrentBlock=BLOCK_COMMENT;

	if( pLinkfile=fopen(tzLinkfile,"rt") )
	{
		while( !feof(pLinkfile) )
		{
			char	tzLine[256];

			fscanf( pLinkfile, "%s\n", tzLine );
			if( tzLine[0]!='#' )
			{
				if( tzLine[0]=='[' && tzLine[strlen(tzLine)-1]==']' )
				{
					if( strcmpi("[objects]",tzLine)==0 )
						CurrentBlock=BLOCK_OBJECTS;
					else if( strcmpi("[output]",tzLine)==0 )
						CurrentBlock=BLOCK_OUTPUT;
					else if( strcmpi("[libraries]",tzLine)==0 )
						CurrentBlock=BLOCK_LIBRARIES;
					else if( strcmpi("[comment]",tzLine)==0 )
						CurrentBlock=BLOCK_COMMENT;
					else
					{
						fclose( pLinkfile );
						sprintf( temptext, "Unknown block '%s'\n", tzLine );
						fatalerror( temptext );
					}
				}
				else
				{
					switch( CurrentBlock )
					{
						case BLOCK_COMMENT:
							break;
						case BLOCK_OBJECTS:
							obj_Readfile( tzLine );
							break;
						case BLOCK_LIBRARIES:
							lib_Readfile( tzLine );
							break;
						case BLOCK_OUTPUT:
							out_Setname( tzLine );
							break;
					}
				}
			}
		}
		fclose( pLinkfile );
	}
	else
	{
		sprintf( temptext, "Unable to find linkfile '%s'\n", tzLinkfile );
		fatalerror( temptext );
	}

}

/*
 * ASJ:
 *
 * A unix-like getopt, but with the names changed to avoid any problems.
 *
 * Stolen from the dumb-frotz source code (I had it just lying around)
 * I'd use GNU's 'official' version, but I want to avoid GPL
 *
 */

static int zoptind = 1;
static int zoptopt = 0;
static char *zoptarg = NULL;
static int zgetopt (int argc, char *argv[], const char *options)
{
    static pos = 1;
    const char *p;
    if (zoptind >= argc || argv[zoptind][0] != '-' || argv[zoptind][1] == 0)
	return EOF;
    zoptopt = argv[zoptind][pos++];
    zoptarg = NULL;
    if (argv[zoptind][pos] == 0)
	{ pos = 1; zoptind++; }
    p = strchr (options, zoptopt);
    if (zoptopt == ':' || p == NULL) {
	fputs ("illegal option -- ", stderr);
	goto error;
    } else if (p[1] == ':')
	if (zoptind >= argc) {
	    fputs ("option requires an argument -- ", stderr);
	    goto error;
	} else {
	    zoptarg = argv[zoptind];
	    if (pos != 1)
		zoptarg += pos;
	    pos = 1; zoptind++;
	}
    return zoptopt;
error:
    fputc (zoptopt, stderr);
    fputc ('\n', stderr);
    return '?';
}/* zgetopt */

/*
 * The main routine
 *
 */

int	main( int argc, char *argv[] )
{
	char c;

	char linkfile[_MAX_PATH];
	linkfile[0] = 0;

	if( argc==0 )
		PrintUsage();

	/* 
	 * ASJ:	Note that I changed the way the it read's the command options to use 
	 *      getopt() style stuff.
	 *
	 */
	while((c = zgetopt(argc, argv, "?hl:m:n:z:s:t:vUu")) != EOF)
	{
		switch(c)
		{
		case '?':
		case 'h':
			PrintUsage();
			break;

		case 'l':
			strcpy(linkfile, zoptarg);
			break;

		case 'm':
			SetMapfileName( zoptarg );
			break;

		case 'n':
			SetSymfileName( zoptarg );
			break;

		case 't':
			if(stricmp(zoptarg,"gb") == 0)
			{
				outputtype=OUTPUT_GBROM;				
				verbose ("Targetting the GB");
			}
			else if(stricmp(zoptarg,"gb-s") == 0)
			{
				outputtype=OUTPUT_GBROM;
				options|=OPT_SMALL;
				verbose ("Targetting the GB (small ROM)");
			}
			else if(stricmp(zoptarg,"p2") == 0)
			{
				outputtype=OUTPUT_PSION2;
				verbose ("Targetting the Psion2");
			}
			else if(stricmp(zoptarg,"ti83") == 0)
			{
				outputtype=OUTPUT_TI83;
				verbose ("Targetting the TI-83");
			}
			else if(stricmp(zoptarg,"ti83+") == 0)
			{
				outputtype=OUTPUT_TI83PLUS;
				verbose ("Targetting the TI-83+");
			}
			else
			{
				sprintf( temptext, "Unknown target: '%s'\n", zoptarg );
				fatalerror( temptext );
			}
			break;

		case 'z':
			if( strlen(zoptarg)<=2 )
			{
				if( strcmp(zoptarg,"?")==0 )
				{
					fillchar=-1;
				}
				else
				{
					int	result;

					result=sscanf( zoptarg, "%x", &fillchar );
					if( !((result==EOF) || (result==1)) )
					{
						fatalerror("Invalid argument for option 'z'\n" );
					}
				}
			}
			else
			{
				fatalerror("Invalid argument for option 'z'\n" );
			}
			break;

		case 's':
			options|=OPT_SMART_C_LINK;
			strcpy( smartlinkstartsymbol, zoptarg );
			sprintf(temptext, "Smat linking starting at %s", smartlinkstartsymbol);
			verbose(temptext);
			break;

		case 'u':
			options|=OPT_UNPROTECTED;
			verbose("Leaving program unprotected");
			break;

		case 'U':
			options|=OPT_UNSQUISHED;
			verbose("Leaving the program unsquished");
			break;

		case 'v':
			usingVerboseMode = 1;
			verbose("Verbose mode is on.");
			break;

		default:
			sprintf( temptext, "Unknown option '%c'\n", c );
			fatalerror( temptext );
			break;
		}
	}

	if( linkfile[0] ) /* make sure the user specified a link file */
	{
		sprintf(temptext, "Linking using '%s'", linkfile);
		verbose(temptext);

		ProcessLinkfile( linkfile );
		AddNeededModules();
		AssignSections();
		CreateSymbolTable();
		Patch();
		Output();
		CloseMapfile();

		printf("Link succesfull!\n\n");
	}
	else
		PrintUsage();

	return( 0 );
}